# -*- coding: utf-8 -*-
"""Work in progress translation of FFC evaluatebasis code to uflacs CNodes format."""

from six import string_types
from ffc.log import error

"""
TODO: Add these to ufc::finite_element:

    /// NEW: Evaluate all reference basis function values at given points X in reference cell
    /// Unflattened shape of values is [num_points][num_dofs][reference_value_size]
    /// Unflattened shape of X is [num_points][tdim]
    virtual void evaluate_reference_basis_values(double * reference_values,
                                                 std::size_t num_points,
                                                 const double * X) const = 0;

    /// NEW: Evaluate all reference basis function derivatives up to order at given points X in reference cell
    /// Unflattened shape of values is [num_points][num_derivatives][num_dofs][reference_value_size]
    /// Unflattened shape of X is [num_points][tdim]
    virtual void evaluate_reference_basis_derivatives(double * reference_values,
                                                      std::size_t num_points,
                                                      std::size_t order,
                                                      const double * X) const = 0;

"""

''' /// FUTURE SPLIT IMPLEMENTATION OF EVALUATE_BASIS:
    /// Evaluate basis function i at given point x in cell
    virtual void evaluate_basis(std::size_t i,
                                double* values,
                                const double* x,
                                const double* coordinate_dofs,
                                int cell_orientation) const;
    /// Evaluate all basis functions at given point x in cell
    virtual void evaluate_basis_all(double* values,
                                    const double* x,
                                    const double* coordinate_dofs,
                                    int cell_orientation) const
    ... and derivatives
    {
      const std::size_t gdim = 3;
      const std::size_t tdim = 2;
      const std::size_t num_points = 1;

      // domain::
      double X[num_points*tdim]; // X[i] -> X[ip*tdim + i]
      compute_reference_coordinates(X, num_points, x, coordinate_dofs, cell_orientation);

      // domain::
      double J[num_points*gdim*tdim]; // J[i,j] -> J[ip*gdim*tdim + i*tdim + j]
      compute_jacobians(J, num_points, X, coordinate_dofs, cell_orientation);

      // domain::
      double detJ[num_points]; // detJ -> detJ[ip]
      compute_jacobian_determinants(detJ, num_points, J);

      // domain::
      double K[num_points*tdim*gdim]; // K[i,j] -> K[ip*tdim*gdim + i*gdim + j]
      compute_jacobian_inverses(K, num_points, J, detJ);

      // domain:: (inverse of compute_reference_coordinates)
      //double x[num_points*gdim]; // x[i] -> x[ip*gdim + i]
      //compute_physical_coordinates(x, num_points, X, K, coordinate_dofs, cell_orientation);

      // domain:: (combining the above)
      //compute_geometry(x, J, detJ, K, num_points, X, coordinate_dofs, cell_orientation);

      // phi[ip*ndofs*rvs + idof*rvs + jcomp]
      double reference_basis_values[num_points*num_dofs*reference_value_size];
      compute_reference_basis(reference_basis_values, num_points, X);

      // phi[ip*nder*ndofs*rvs + iderivative*ndofs*rvs + idof*rvs + jcomp]
      double reference_basis_derivatives[num_points*num_derivatives*num_dofs*reference_value_size];
      compute_reference_basis_derivatives(reference_basis_derivatives, derivative_order, num_points, X);

      double physical_basis_values[num_points*num_dofs*value_size]; // phi -> phi[ip*ndofs*pvs + idof*pvs + icomp]
      compute_physical_basis[_derivatives](physical_basis_values, num_points, reference_basis_values, J, detJ, K);
    }
'''

import math

def generate_evaluate_reference_basis(L, data):
    """Generate code to evaluate element basisfunctions at an arbitrary point on the reference element.

    The value(s) of the basisfunction is/are computed as in FIAT as
    the dot product of the coefficients (computed at compile time)
    and basisvalues which are dependent on the coordinate and thus
    have to be computed at run time.

    The function should work for all elements supported by FIAT, but
    it remains untested for tensor valued elements.

    This code is adapted from code in FFC which computed the basis
    from physical coordinates, and also to use UFLACS utilities.

    The FFC code has a comment "From FIAT_NEW.polynomial_set.tabulate()".
    """
    # Cutoff for feature to disable generation of this code (consider removing after benchmarking final result)
    if isinstance(data, string_types):
        return L.Throw("evaluate_reference_basis: %s" % data)

    # Get some known dimensions
    element_cellname = data["cellname"]
    #gdim = data["geometric_dimension"]
    tdim = data["topological_dimension"]
    reference_value_size = data["reference_value_size"]
    num_dofs = len(data["dofs_data"])

    # Input geometry
    num_points = L.Symbol("num_points")
    X = L.Symbol("X")

    # Output values
    reference_values = L.Symbol("reference_values")
    ref_values = L.FlattenedArray(reference_values, dims=(num_points, num_dofs, reference_value_size))

    # NB! This symbol refers to the FIAT reference coordinate!
    Y = L.Symbol("Y")

    # Loop indices
    ip = L.Symbol("ip")
    k = L.Symbol("k")
    c = L.Symbol("c")
    r = L.Symbol("r")


    # Generate code with static tables of expansion coefficients
    tables_code = []
    coefficients_for_dof = []
    for idof, dof_data in enumerate(data["dofs_data"]):
        num_components = dof_data["num_components"]
        num_members = dof_data["num_expansion_members"]
        fiat_coefficients = dof_data["coeffs"]

        # TODO: Check if any fiat_coefficients tables in expansion_coefficients are equal and reuse instead of declaring new.

        # Create separate variable name for coefficients table for each dof
        coefficients = L.Symbol("coefficients%d" % idof)

        # Create static table with expansion coefficients computed by FIAT compile time.
        tables_code += [L.ArrayDecl("static const double", coefficients,
                                    (num_components, num_members), values=fiat_coefficients)]

        # Store symbol reference for this dof
        coefficients_for_dof.append(coefficients)


    # Reset values[:] to 0
    reset_values_code = [
        L.ForRange(k, 0, num_points*num_dofs*reference_value_size, body=
            L.Assign(reference_values[k], 0.0))
        ]


    # Mapping from UFC reference cell coordinate X to FIAT reference cell coordinate Y
    fiat_coordinate_mapping = L.ArrayDecl("const double", Y, (tdim,),
                                          values=[2.0*X[ip*tdim + jj]-1.0 for jj in range(tdim)])


    # Generate code to compute tables of basisvalues
    basisvalues_code = []
    basisvalues_for_degree = {}
    for idof, dof_data in enumerate(data["dofs_data"]):
        embedded_degree = dof_data["embedded_degree"]
        num_members = dof_data["num_expansion_members"]

        if embedded_degree not in basisvalues_for_degree:
            basisvalues = L.Symbol("basisvalues%d" % embedded_degree)
            bfcode = _generate_compute_basisvalues(L, basisvalues, Y, element_cellname, embedded_degree, num_members)
            basisvalues_code += [L.StatementList(bfcode)]

            # Store symbol reference for this degree
            basisvalues_for_degree[embedded_degree] = basisvalues


    # Accumulate products of basisvalues and coefficients into values
    accumulation_code = []
    for idof, dof_data in enumerate(data["dofs_data"]):
        embedded_degree = dof_data["embedded_degree"]
        num_components = dof_data["num_components"]
        num_members = dof_data["num_expansion_members"]

        # In ffc representation, this is extracted per dof (but will coincide for some dofs of piola mapped elements):
        reference_offset = dof_data["reference_offset"]

        # Select the right basisvalues for this dof
        basisvalues = basisvalues_for_degree[embedded_degree]

        # Select the right coefficients for this dof
        coefficients = coefficients_for_dof[idof]

        # Generate basis accumulation loop
        if num_components > 1:
            accumulation_code += [
                L.ForRange(c, 0, num_components, body=
                    L.ForRange(r, 0, num_members, body=
                        L.AssignAdd(ref_values[ip, idof, reference_offset + c],
                                    coefficients[c,r] * basisvalues[r])))
                ]
        elif num_members > 1:
            accumulation_code += [
                L.ForRange(r, 0, num_members, body=
                    L.AssignAdd(ref_values[ip, idof, reference_offset], coefficients[0, r] * basisvalues[r]))
                ]
        else:
            accumulation_code += [
                L.AssignAdd(ref_values[ip, idof, reference_offset], coefficients[0, 0] * basisvalues[0])
                ]

        # FIXME: Move this mapping to its own ufc function e.g. finite_element::apply_element_mapping(reference_values, J, K)
        #code += _generate_apply_mapping_to_computed_values(L, dof_data) # Only works for affine (no-op)

    # Stitch it all together
    code = L.StatementList(
        tables_code +
        reset_values_code +
        [L.ForRange(ip, 0, num_points, body=L.StatementList([
            L.Comment("Map from UFC reference coordinate X to FIAT reference coordinate Y"),
            fiat_coordinate_mapping,
            L.Comment("Compute basisvalues for each relevant embedded degree"),
            basisvalues_code,
            L.Comment("Accumulate products of coefficients and basisvalues"),
            accumulation_code,
            ]))])
    return code


def _generate_compute_basisvalues(L, basisvalues, Y, element_cellname, embedded_degree, num_members):
    """From FIAT_NEW.expansions."""

    # Branch off to cell specific implementations
    if element_cellname == "interval":
        code = _generate_compute_interval_basisvalues(L, basisvalues, Y, embedded_degree, num_members)
    elif element_cellname == "triangle":
        code = _generate_compute_triangle_basisvalues(L, basisvalues, Y, embedded_degree, num_members)
    elif element_cellname == "tetrahedron":
        code = _generate_compute_tetrahedron_basisvalues(L, basisvalues, Y, embedded_degree, num_members)
    else:
        error()

    return code

def _jrc(a, b, n):
    an = float((2*n+1+a+b) * (2*n+2+a+b))   / float(2*(n+1) * (n+1+a+b))
    bn = float((a*a-b*b) * (2*n+1+a+b))     / float(2*(n+1) * (2*n+a+b) * (n+1+a+b))
    cn = float((n+a) * (n+b) * (2*n+2+a+b)) / float((n+1) * (n+1+a+b) * (2*n+a+b))
    return (an, bn, cn)

def _generate_compute_interval_basisvalues(L, basisvalues, Y, embedded_degree, num_members):
    # FIAT_NEW.expansions.LineExpansionSet.

    # Create zero-initialized array for with basisvalues
    code = [L.ArrayDecl("double", basisvalues, (num_members,), values=0)]

    # FIAT_NEW.jacobi.eval_jacobi_batch(a,b,n,xs)
    # for ii in range(result.shape[1]):
    #    result[0,ii] = 1.0 + xs[ii,0] - xs[ii,0]
    # The initial value basisvalues[0] is always 1.0
    code += [L.Assign(basisvalues[0], 1.0)]

    if embedded_degree > 0:
        # FIAT_NEW.jacobi.eval_jacobi_batch(a,b,n,xs).
        # result[1,:] = 0.5 * ( a - b + ( a + b + 2.0 ) * xsnew )
        # The initial value basisvalues[1] is always Y[0]
        code += [L.Assign(basisvalues[1], Y[0])]

    # Only active if embedded_degree > 1.
    for r in range(2, embedded_degree+1):
        # FIAT_NEW.jacobi.eval_jacobi_batch(a,b,n,xs).
        # apb = a + b (equal to 0 because of function arguments)
        # for k in range(2,n+1):
        #    a1 = 2.0 * k * ( k + apb ) * ( 2.0 * k + apb - 2.0 )
        #    a2 = ( 2.0 * k + apb - 1.0 ) * ( a * a - b * b )
        #    a3 = ( 2.0 * k + apb - 2.0 )  \
        #        * ( 2.0 * k + apb - 1.0 ) \
        #        * ( 2.0 * k + apb )
        #    a4 = 2.0 * ( k + a - 1.0 ) * ( k + b - 1.0 ) \
        #        * ( 2.0 * k + apb )
        #    a2 = a2 / a1
        #    a3 = a3 / a1
        #    a4 = a4 / a1
        #    result[k,:] = ( a2 + a3 * xsnew ) * result[k-1,:] \
        #        - a4 * result[k-2,:]

        # The below implements the above (with a = b = apb = 0)
        a1 = float(2*r*r*(2*r - 2))
        a3 = ((2*r - 2)*(2*r - 1)*(2*r)) / a1
        a4 = (2*(r - 1)*(r - 1)*(2*r)) / a1
        value = (Y[0] * (a3 * basisvalues[r-1])) - a4*basisvalues[r-2]
        code += [L.Assign(basisvalues[r], value)]

    # FIAT_NEW code
    # results = numpy.zeros( ( n+1 , len(pts) ) , type( pts[0][0] ) )
    # for k in range( n + 1 ):
    #    results[k,:] = psitilde_as[k,:] * math.sqrt( k + 0.5 )

    # Scale values
    p = L.Symbol("p")
    code += [L.ForRange(p, 0, embedded_degree + 1,
                        body=L.AssignMul(basisvalues[p], L.Sqrt(0.5 + p)))]
    return code

def _generate_compute_triangle_basisvalues(L, basisvalues, Y, embedded_degree, num_members):
    # FIAT_NEW.expansions.TriangleExpansionSet.

    def _idx2d(p, q):
        return (p + q)*(p + q + 1)//2 + q

    # Create zero-initialized array for with basisvalues
    code = [L.ArrayDecl("double", basisvalues, (num_members,), values=0)]

    # Compute helper factors
    # FIAT_NEW code
    # f1 = (1.0+2*x+y)/2.0
    # f2 = (1.0 - y) / 2.0
    # f3 = f2**2

    # The initial value basisvalues 0 is always 1.0.
    # FIAT_NEW code
    # for ii in range( results.shape[1] ):
    #    results[0,ii] = 1.0 + apts[ii,0]-apts[ii,0]+apts[ii,1]-apts[ii,1]
    code += [L.Assign(basisvalues[0], 1.0)]

    # Only continue if the embedded degree is larger than zero.
    if embedded_degree == 0:
        return code

    # The initial value of basisfunction 1 is equal to f1.
    # FIAT_NEW code
    # results[idx(1,0),:] = f1
    f1 = L.Symbol("tmp1_%d" % embedded_degree)
    code += [L.VariableDecl("const double", f1, (1.0 + 2.0*Y[0] + Y[1]) / 2.0)]
    code += [L.Assign(basisvalues[1], f1)]

    # NOTE: KBO: The order of the loops is VERY IMPORTANT!!

    # FIAT_NEW code (loop 1 in FIAT)
    # for p in range(1,n):
    #    a = (2.0*p+1)/(1.0+p)
    #    b = p / (p+1.0)
    #    results[idx(p+1,0)] = a * f1 * results[idx(p,0),:] \
    #        - b * f3 *results[idx(p-1,0),:]
    # Only active if embedded_degree > 1.
    if embedded_degree > 1:
        f2 = L.Symbol("tmp2_%d" % embedded_degree)
        f3 = L.Symbol("tmp3_%d" % embedded_degree)
        code += [L.VariableDecl("const double", f2, (1.0 - Y[1]) / 2.0)]
        code += [L.VariableDecl("const double", f3, f2*f2)]
    for r in range(1, embedded_degree):
        rr = _idx2d((r + 1), 0)
        ss = _idx2d(r, 0)
        tt = _idx2d((r - 1), 0)
        A = (2*r + 1.0)/(r + 1)
        B = r/(1.0 + r)
        v1 = A * f1 * basisvalues[ss]
        v2 = B * f3 * basisvalues[tt]
        value = v1 - v2
        code += [L.Assign(basisvalues[rr], value)]

    # FIAT_NEW code (loop 2 in FIAT).
    # for p in range(n):
    #    results[idx(p,1),:] = 0.5 * (1+2.0*p+(3.0+2.0*p)*y) \
    #        * results[idx(p,0)]
    for r in range(0, embedded_degree):
        # (p+q)*(p+q+1)//2 + q
        rr = _idx2d(r, 1)
        ss = _idx2d(r, 0)
        A = 0.5*(1 + 2*r)
        B = 0.5*(3 + 2*r)
        C = A + (B * Y[1])
        value = C * basisvalues[ss]
        code += [L.Assign(basisvalues[rr], value)]

    # Only active if embedded_degree > 1.
    # FIAT_NEW code (loop 3 in FIAT).
    # for p in range(n-1):
    #    for q in range(1,n-p):
    #        (a1,a2,a3) = jrc(2*p+1,0,q)
    #        results[idx(p,q+1),:] \
    #            = ( a1 * y + a2 ) * results[idx(p,q)] \
    #            - a3 * results[idx(p,q-1)]
    # Only active if embedded_degree > 1.
    for r in range(0, embedded_degree - 1):
        for s in range(1, embedded_degree - r):
            rr = _idx2d(r, (s + 1))
            ss = _idx2d(r, s)
            tt = _idx2d(r, s - 1)
            A, B, C = _jrc(2*r + 1, 0, s)
            value = (B + A*Y[1])*basisvalues[ss] - C*basisvalues[tt]
            code += [L.Assign(basisvalues[rr], value)]

    # FIAT_NEW code (loop 4 in FIAT).
    # for p in range(n+1):
    #    for q in range(n-p+1):
    #        results[idx(p,q),:] *= math.sqrt((p+0.5)*(p+q+1.0))
    for r in range(0, embedded_degree + 1):
        for s in range(0, embedded_degree + 1 - r):
            rr = _idx2d(r, s)
            A = (r + 0.5)*(r + s + 1)
            code += [L.AssignMul(basisvalues[rr], math.sqrt(A))]

    return code

def _generate_compute_tetrahedron_basisvalues(L, basisvalues, Y, embedded_degree, num_members):
    # FIAT_NEW.expansions.TetrahedronExpansionSet.

    # FIAT_NEW code (compute index function) TetrahedronExpansionSet.
    # def idx(p,q,r):
    #     return (p+q+r)*(p+q+r+1)*(p+q+r+2)//6 + (q+r)*(q+r+1)//2 + r
    def _idx3d(p, q, r):
        return (p+q+r)*(p+q+r+1)*(p+q+r+2)//6 + (q+r)*(q+r+1)//2 + r

    # Compute helper factors.
    # FIAT_NEW code
    # factor1 = 0.5 * ( 2.0 + 2.0*x + y + z )
    # factor2 = (0.5*(y+z))**2
    # factor3 = 0.5 * ( 1 + 2.0 * y + z )
    # factor4 = 0.5 * ( 1 - z )
    # factor5 = factor4 ** 2

    # Create zero-initialized array for with basisvalues
    code = [L.ArrayDecl("double", basisvalues, (num_members,), values=0)]

    # The initial value basisvalues 0 is always 1.0.
    # FIAT_NEW code
    # for ii in range( results.shape[1] ):
    #    results[0,ii] = 1.0 + apts[ii,0]-apts[ii,0]+apts[ii,1]-apts[ii,1]
    code += [L.Assign(basisvalues[0], 1.0)]

    # Only continue if the embedded degree is larger than zero.
    if embedded_degree == 0:
        return code

    # The initial value of basisfunction 1 is equal to f1.
    # FIAT_NEW code
    # results[idx(1,0),:] = f1
    f1 = L.Symbol("tmp1_%d" % embedded_degree)
    code += [L.VariableDecl("const double", f1, 0.5*(2.0 + 2.0*Y[0] + Y[1] + Y[2]))]
    code += [L.Assign(basisvalues[1], f1)]

    # NOTE: KBO: The order of the loops is VERY IMPORTANT!!

    # FIAT_NEW code (loop 1 in FIAT).
    # for p in range(1,n):
    #    a1 = ( 2.0 * p + 1.0 ) / ( p + 1.0 )
    #    a2 = p / (p + 1.0)
    #    results[idx(p+1,0,0)] = a1 * factor1 * results[idx(p,0,0)] \
    #        -a2 * factor2 * results[ idx(p-1,0,0) ]
    # Only active if embedded_degree > 1.
    if embedded_degree > 1:
        f2 = L.Symbol("tmp2_%d" % embedded_degree)
        code += [L.VariableDecl("const double", f2, 0.25*(Y[1] + Y[2])*(Y[1] + Y[2]))]
    for r in range(1, embedded_degree):
        rr = _idx3d((r + 1), 0, 0)
        ss = _idx3d(r, 0, 0)
        tt = _idx3d((r - 1), 0, 0)
        A = (2*r + 1.0)/(r + 1)
        B = r/(r + 1.0)
        value = (A * f1 * basisvalues[ss]) - (B * f2 * basisvalues[tt])
        code += [L.Assign(basisvalues[rr], value)]

    # FIAT_NEW code (loop 2 in FIAT).
    # q = 1
    # for p in range(0,n):
    #    results[idx(p,1,0)] = results[idx(p,0,0)] \
    #        * ( p * (1.0 + y) + ( 2.0 + 3.0 * y + z ) / 2 )
    for r in range(0, embedded_degree):
        rr = _idx3d(r, 1, 0)
        ss = _idx3d(r, 0, 0)
        term0 = 0.5 * (2.0 + 3.0*Y[1] + Y[2])
        term1 = float(r) * (1.0 + Y[1])
        value = (term0 + term1) * basisvalues[ss]
        code += [L.Assign(basisvalues[rr], value)]

    # FIAT_NEW code (loop 3 in FIAT).
    # for p in range(0,n-1):
    #    for q in range(1,n-p):
    #        (aq,bq,cq) = jrc(2*p+1,0,q)
    #        qmcoeff = aq * factor3 + bq * factor4
    #        qm1coeff = cq * factor5
    #        results[idx(p,q+1,0)] = qmcoeff * results[idx(p,q,0)] \
    #            - qm1coeff * results[idx(p,q-1,0)]
    # Only active if embedded_degree > 1.
    if embedded_degree > 1:
        f3 = L.Symbol("tmp3_%d" % embedded_degree)
        f4 = L.Symbol("tmp4_%d" % embedded_degree)
        f5 = L.Symbol("tmp5_%d" % embedded_degree)
        code += [L.VariableDecl("const double", f3, 0.5*(1.0 + 2.0*Y[1] + Y[2]))]
        code += [L.VariableDecl("const double", f4, 0.5*(1.0 - Y[2]))]
        code += [L.VariableDecl("const double", f5, f4*f4)]
    for r in range(0, embedded_degree - 1):
        for s in range(1, embedded_degree - r):
            rr = _idx3d(r, (s + 1), 0)
            ss = _idx3d(r, s, 0)
            tt = _idx3d(r, s - 1, 0)
            (A, B, C) = _jrc(2*r + 1, 0, s)
            term0 = ((A*f3) + (B*f4)) * basisvalues[ss]
            term1 = C * f5 * basisvalues[tt]
            value = term0 - term1
            code += [L.Assign(basisvalues[rr], value)]

    # FIAT_NEW code (loop 4 in FIAT).
    # now handle r=1
    # for p in range(n):
    #    for q in range(n-p):
    #        results[idx(p,q,1)] = results[idx(p,q,0)] \
    #            * ( 1.0 + p + q + ( 2.0 + q + p ) * z )
    for r in range(0, embedded_degree):
        for s in range(0, embedded_degree - r):
            rr = _idx3d(r, s, 1)
            ss = _idx3d(r, s, 0)
            A = (float(2 + r + s) * Y[2]) + float(1 + r + s)
            value = A * basisvalues[ss]
            code += [L.Assign(basisvalues[rr], value)]

    # FIAT_NEW code (loop 5 in FIAT).
    # general r by recurrence
    # for p in range(n-1):
    #     for q in range(0,n-p-1):
    #         for r in range(1,n-p-q):
    #             ar,br,cr = jrc(2*p+2*q+2,0,r)
    #             results[idx(p,q,r+1)] = \
    #                         (ar * z + br) * results[idx(p,q,r) ] \
    #                         - cr * results[idx(p,q,r-1) ]
    # Only active if embedded_degree > 1.
    for r in range(0, embedded_degree - 1):
        for s in range(0, embedded_degree - r - 1):
            for t in range(1, embedded_degree - r - s):
                rr = _idx3d(r, s, ( t + 1))
                ss = _idx3d(r, s, t)
                tt = _idx3d(r, s, t - 1)
                (A, B, C) = _jrc(2*r + 2*s + 2, 0, t)
                az_b = B + A*Y[2]
                value = (az_b*basisvalues[ss]) - (C*basisvalues[tt])
                code += [L.Assign(basisvalues[rr], value)]

    # FIAT_NEW code (loop 6 in FIAT).
    # for p in range(n+1):
    #    for q in range(n-p+1):
    #        for r in range(n-p-q+1):
    #            results[idx(p,q,r)] *= math.sqrt((p+0.5)*(p+q+1.0)*(p+q+r+1.5))
    for r in range(0, embedded_degree + 1):
        for s in range(0, embedded_degree - r + 1):
            for t in range(0, embedded_degree - r - s + 1):
                rr = _idx3d(r, s, t)
                A = (r + 0.5)*(r + s + 1)*(r + s + t + 1.5)
                code += [L.AssignMul(basisvalues[rr], math.sqrt(A))]

    return code


def _generate_apply_mapping_to_computed_values(L, dof_data):
    mapping = dof_data["mapping"]
    num_components = dof_data["num_components"]
    reference_offset = dof_data["reference_offset"]
    physical_offset = dof_data["physical_offset"]

    physical_values[num_points][num_dofs][physical_value_size]
    reference_values[num_points][num_dofs][reference_value_size]

    code = []

    # FIXME: Define values numbering
    if mapping == "affine":
        # Just copy values
        if num_components == 1:
            code += [
                L.ForRange(ip, 0, num_points, body=
                    L.Assign(physical_values[ip][physical_offset],
                             reference_values[ip][reference_offset])),
                ]
        else:
            code += [
                L.ForRange(ip, 0, num_points, body=
                    L.ForRange(k, 0, num_components, body=
                        L.Assign(physical_values[ip][physical_offset + k],
                                 reference_values[ip][reference_offset + k]))),
                ]
        return code

    else:
        fixme

    return code

def __ffc_implementation_of__generate_apply_mapping_to_computed_values(L):
    # Apply transformation if applicable.
    mapping = dof_data["mapping"]
    num_components = dof_data["num_components"]
    reference_offset = dof_data["reference_offset"]
    physical_offset = dof_data["physical_offset"]

    if mapping == "affine":
        pass

    elif mapping == "contravariant piola":
        code += ["", f_comment("Using contravariant Piola transform to map values back to the physical element")]

        # Get temporary values before mapping.
        code += [f_const_float(f_tmp_ref(i), f_component(f_values, i + offset))
                 for i in range(num_components)]

        # Create names for inner product.
        basis_col = [f_tmp_ref(j) for j in range(tdim)]
        for i in range(gdim):
            # Create Jacobian.
            jacobian_row = [f_trans("J", i, j, gdim, tdim, None) for j in range(tdim)]
            # Create inner product and multiply by inverse of Jacobian.
            inner = f_group(f_inner(jacobian_row, basis_col)) # sum_j J[i,j], values[offset + j])
            value = f_mul([f_inv(f_detJ(None)), inner])
            name = f_component(f_values, i + offset)
            # FIXME: This is writing values[offset+:] = M[:,:] * values[offset+:],
            #        i.e. offset must be physical (unless there's a bug).
            #        We want to use reference offset for evaluate_reference_basis,
            #        and to make this mapping read from one array using reference offset
            #        and write to another array using physical offset.
            code += [f_assign(name, value)]

    elif mapping == "covariant piola":
        code += ["", f_comment("Using covariant Piola transform to map values back to the physical element")]
        # Get temporary values before mapping.
        code += [f_const_float(f_tmp_ref(i), f_component(f_values, i + offset))
                 for i in range(num_components)]
        # Create names for inner product.
        tdim = data["topological_dimension"]
        gdim = data["geometric_dimension"]
        basis_col = [f_tmp_ref(j) for j in range(tdim)]
        for i in range(gdim):
            # Create inverse of Jacobian.
            inv_jacobian_column = [f_trans("JINV", j, i, tdim, gdim, None) for j in range(tdim)]

            # Create inner product of basis values and inverse of Jacobian.
            value = f_group(f_inner(inv_jacobian_column, basis_col))
            name = f_component(f_values, i + offset)
            code += [f_assign(name, value)]

    elif mapping == "double covariant piola":
        code += ["", f_comment("Using double covariant Piola transform to map values back to the physical element")]
        # Get temporary values before mapping.
        code += [f_const_float(f_tmp_ref(i), f_component(f_values, i + offset))
                 for i in range(num_components)]
        # Create names for inner product.
        tdim = data["topological_dimension"]
        gdim = data["geometric_dimension"]
        basis_col = [f_tmp_ref(j) for j in range(num_components)]
        for p in range(num_components):
            # unflatten the indices
            i = p // tdim
            l = p % tdim
            # g_il = K_ji G_jk K_kl
            value = f_group(f_inner(
                [f_inner([f_trans("JINV", j, i, tdim, gdim, None)
                          for j in range(tdim)],
                         [basis_col[j * tdim + k] for j in range(tdim)])
                 for k in range(tdim)],
                [f_trans("JINV", k, l, tdim, gdim, None)
                 for k in range(tdim)]))
            name = f_component(f_values, p + offset)
            code += [f_assign(name, value)]

    elif mapping == "double contravariant piola":
        code += ["", f_comment("Pullback of a matrix-valued funciton as contravariant 2-tensor mapping values back to the physical element")]
        # Get temporary values before mapping.
        code += [f_const_float(f_tmp_ref(i), f_component(f_values, i + offset))
                 for i in range(num_components)]
        # Create names for inner product.
        tdim = data["topological_dimension"]
        gdim = data["geometric_dimension"]
        basis_col = [f_tmp_ref(j) for j in range(num_components)]
        for p in range(num_components):
            # unflatten the indices
            i = p // tdim
            l = p % tdim
            # g_il = (detJ)^(-2) J_ij G_jk J_lk
            value = f_group(f_inner(
                [f_inner([f_trans("J", i, j, tdim, gdim, None)
                          for j in range(tdim)],
                         [basis_col[j * tdim + k] for j in range(tdim)])
                 for k in range(tdim)],
                [f_trans("J", l, k, tdim, gdim, None)
                 for k in range(tdim)]))
            value = f_mul([f_inv(f_detJ(None)), f_inv(f_detJ(None)), value])
            name = f_component(f_values, p + offset)
            code += [f_assign(name, value)]

    else:
        error("Unknown mapping: %s" % mapping)



# -*- coding: utf-8 -*-
"""
Tests of graph representation of expressions.
"""

from __future__ import print_function

from six.moves import xrange as range
from ufl import *
from ufl import product
from ufl.permutation import compute_indices

from ffc.uflacs.analysis.graph import build_graph
from ffc.uflacs.analysis.graph_rebuild import rebuild_expression_from_graph
# from ffc.uflacs.analysis.graph_rebuild import rebuild_scalar_e2i
# from ffc.uflacs.analysis.dependencies import (compute_dependencies,
#                                                mark_active,
#                                                mark_image)
# from ffc.uflacs.analysis.graph_ssa import (mark_partitions,
#                                       compute_dependency_count,
#                                       invert_dependencies,
#                                       default_cache_score_policy,
#                                       compute_cache_scores,
#                                       allocate_registers)

from operator import eq as equal


def test_graph_algorithm_allocates_correct_number_of_symbols():
    U = FiniteElement("CG", triangle, 1)
    V = VectorElement("CG", triangle, 1)
    W = TensorElement("CG", triangle, 1)
    u = Coefficient(U)
    v = Coefficient(V)
    w = Coefficient(W)

    # Testing some scalar expressions
    expr = u
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 1
    assert G.total_unique_symbols == 1

    expr = u**2
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 3
    assert G.total_unique_symbols == 3

    expr = u**2 / 2
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 4
    assert G.total_unique_symbols == 4

    expr = dot(v, v) / 2
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 5
    assert G.total_unique_symbols == 5

    # Testing Indexed
    expr = v[i] * v[i]
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 2 + 2 + 2 + 1
    assert G.total_unique_symbols == 2 + 2 + 1

    # Reusing symbols for indexed with different ordering
    # Note that two index sums are created, giving 2+1 symbols
    expr = w[i, j] * w[j, i]
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 4 + 4 + 4 + 4 + 2 + 1
    assert G.total_unique_symbols == 4 + 4 + 2 + 1

    # Testing ComponentTensor
    expr = dot(as_vector(2 * v[i], i), v)
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 2 + 1 + 2 + 2 + 2 + 1
    assert G.total_unique_symbols == 2 + 1 + 2 + 1

    expr = dot(v + 2 * v, v)
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 2 + 1 + 2 + 2 + 2 + 2 + 1
    assert G.total_unique_symbols == 2 + 1 + 2 + 2 + 1

    expr = outer(v, v)[i, j] * outer(v, v)[j, i]
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 21  # 2+4+4+4 + 4+2+1
    assert G.total_unique_symbols == 13  # 2+4+4 + 2+1

    # Testing tensor/scalar
    expr = as_ufl(2)
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 1
    assert G.total_unique_symbols == 1

    expr = v
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 2
    assert G.total_unique_symbols == 2

    expr = outer(v, v)
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 2 + 4
    assert G.total_unique_symbols == 2 + 4

    expr = as_tensor(v[i] * v[j], (i, j))
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 2 + 2 + 2 + 4 + 4
    assert G.total_unique_symbols == 2 + 4

    expr = as_tensor(v[i] * v[j] / 2, (i, j))
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 2 + 2 + 2 + 4 + 4 + 4 + 1
    assert G.total_unique_symbols == 2 + 1 + 4 + 4

    expr = outer(v, v) / 2  # converted to the tensor notation above
    G = build_graph([expr], DEBUG=0)
    assert G.V_symbols.num_elements == 2 + 2 + 2 + 4 + 4 + 4 + 1
    assert G.total_unique_symbols == 2 + 1 + 4 + 4


def test_rebuild_expression_from_graph_basic_scalar_expressions():
    U = FiniteElement("CG", triangle, 1)
    V = VectorElement("CG", triangle, 1)
    W = TensorElement("CG", triangle, 1)
    u = Coefficient(U)
    v = Coefficient(V)
    w = Coefficient(W)

    # ... Literals are reproduced
    literals = [as_ufl(0), as_ufl(1), as_ufl(3.14)]
    for v1 in literals:
        G = build_graph([v1])
        v2 = rebuild_expression_from_graph(G)
        assert v1 == v2

    v1 = u
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    assert v1 == v2

    # ... Simple operators are reproduced
    for v1 in [2 + u, u + u, u * u, u * 2, u**2, u**u, u / u, sin(u)]:
        G = build_graph([v1])
        v2 = rebuild_expression_from_graph(G)
        assert v1 == v2


def test_rebuild_expression_from_graph_on_products_with_indices():
    U = FiniteElement("CG", triangle, 1)
    V = VectorElement("CG", triangle, 1)
    W = TensorElement("CG", triangle, 1)
    u = Coefficient(U)
    v = Coefficient(V)
    w = Coefficient(W)
    i, j, k, l = indices(4)

    # Test fixed index
    fixed = [u * v[0], v[1] * v[0], w[0, 1] * w[0, 0]]
    for v1 in fixed:
        G = build_graph([v1])
        v2 = rebuild_expression_from_graph(G)
        assert v1 == v2

    # Test simple repeated index
    v1 = v[i] * v[i]
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = v[0] * v[0] + v[1] * v[1]
    assert ve == v2

    # Test double repeated index
    v1 = w[i, j] * w[j, i]
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = (w[1, 1] * w[1, 1] + w[1, 0] * w[0, 1]) + (w[0, 1] * w[1, 0] + w[0, 0] * w[0, 0])
    if 0:
        print()
        print(v1)
        print(ve)
        print(v2)
        print()
    assert ve == v2

    # Test mix of repeated and non-repeated index
    v1 = (w[i, j] * w[j, 0] + v[i]) * v[i]
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = ((w[0, 0] * w[0, 0] + w[0, 1] * w[1, 0] + v[0]) * v[0]
          + (w[1, 0] * w[0, 0] + w[1, 1] * w[1, 0] + v[1]) * v[1])
    assert ve == v2


def test_rebuild_expression_from_graph_basic_tensor_expressions():
    U = FiniteElement("CG", triangle, 1)
    V = VectorElement("CG", triangle, 1)
    W = TensorElement("CG", triangle, 1)
    u = Coefficient(U)
    v = Coefficient(V)
    vb = Coefficient(V)
    w = Coefficient(W)
    wb = Coefficient(W)

    # Single vector
    v1 = v
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    assert as_vector((v1[0], v1[1])) == v2

    # Single tensor
    v1 = w
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    assert as_vector((v1[0, 0], v1[0, 1], v1[1, 0], v1[1, 1])) == v2

    # Vector sum
    v1 = v + v
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = as_vector((v[0] + v[0], v[1] + v[1]))
    assert ve == v2

    v1 = v + vb
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = as_vector((v[0] + vb[0], v[1] + vb[1]))
    assert ve == v2

    # Tensor sum
    v1 = w + w
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = as_vector((w[0, 0] + w[0, 0], w[0, 1] + w[0, 1], w[1, 0] + w[1, 0], w[1, 1] + w[1, 1]))
    assert ve == v2

    v1 = w + wb
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = as_vector((w[0, 0] + wb[0, 0], w[0, 1] + wb[0, 1], w[1, 0] + wb[1, 0], w[1, 1] + wb[1, 1]))
    assert ve == v2

    # Scalar-vector product
    v1 = u * v
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = as_vector((u * v[0], u * v[1]))
    assert ve == v2

    # Scalar-tensor product
    v1 = u * w
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = as_vector((u * w[0, 0], u * w[0, 1], u * w[1, 0], u * w[1, 1]))
    assert ve == v2

    # Vector-vector index based inner product
    v1 = v[i] * v[i]
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = v[0] * v[0] + v[1] * v[1]
    assert ve == v2

    v1 = v[i] * vb[i]
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = v[0] * vb[0] + v[1] * vb[1]
    assert ve == v2

    # Tensor-tensor index based transposed inner product
    v1 = w[i, j] * w[j, i]
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = (w[0, 0] * w[0, 0] + w[0, 1] * w[1, 0]) \
        + (w[1, 0] * w[0, 1] + w[1, 1] * w[1, 1])
    assert ve == v2

    v1 = w[i, j] * wb[j, i]
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = (w[0, 0] * wb[0, 0] + w[1, 0] * wb[0, 1]) \
        + (w[0, 1] * wb[1, 0] + w[1, 1] * wb[1, 1])
    assert ve == v2

    # Vector/scalar division
    v1 = v / u
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = as_vector((v[0] / u, v[1] / u))
    assert ve == v2

    # Tensor/scalar division
    v1 = w / u
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    ve = as_vector((w[0, 0] / u, w[0, 1] / u, w[1, 0] / u, w[1, 1] / u))
    assert ve == v2

    # FIXME: Write more tests to discover bugs in ReconstructScalarSubexpressions.element_wise*

    # assert False

# Compounds not implemented, not expecting to do this anytime soon


def xtest_rebuild_expression_from_graph_on_compounds():
    U = FiniteElement("CG", triangle, 1)
    V = VectorElement("CG", triangle, 1)
    W = TensorElement("CG", triangle, 1)
    u = Coefficient(U)
    v = Coefficient(V)
    w = Coefficient(W)

    v1 = dot(v, v)
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)

    v1 = outer(v, v)
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)

    v1 = outer(v, v)[i, j] * outer(v, v)[j, i]
    G = build_graph([v1])
    v2 = rebuild_expression_from_graph(G)
    # print v1
    # print v2
    # FIXME: Assert something


def test_flattening_of_tensor_valued_expression_symbols():
    # from ffc.uflacs.analysis.graph import foo
    def flatten_expression_symbols(v, vsyms, opsyms):
        sh = v.ufl_shape
        if sh == ():
            assert len(vsyms) == 1
            if not opsyms:
                res = [(v, vsyms[0], ())]
            else:
                assert len(opsyms[0]) == 1
                res = [(v, vsyms[0], tuple(syms[0] for syms in opsyms))]
        else:
            res = []
            if isinstance(v, ufl.classes.Sum):
                for i in range(len(vsyms)):
                    u = None  # sum of component i for syms in opsyms
                    res += (u, vsyms[i], tuple(syms[i] for syms in opsyms))
        return res

    v = as_ufl(1)
    vsyms = (0,)
    opsyms = ()
    res = flatten_expression_symbols(v, vsyms, opsyms)
    assert res == [(v, 0, ())]



#!/usr/bin/env py.test
# -*- coding: utf-8 -*-

"""
These tests check that the mock implementations
of ufc cells are properly constructed. The mock cells
are not actual subclasses of ufc::cell, instead using
sufficiently large fixed size arrays to avoid inconvenient
memory allocation code in the unit tests. The tests here
can be seen as a demonstration of how to use the mock cells.
"""

def test_mock_interval(gtest):
    pre = """
    mock_cell mc;
    mc.fill_reference_interval(1);
    double * coordinate_dofs = mc.coordinate_dofs;
    """

    post = """
    ASSERT_EQ(coordinate_dofs[0*mc.geometric_dimension + 0], 0.0);
    ASSERT_EQ(coordinate_dofs[1*mc.geometric_dimension + 0], 2.0);
    """

    code = """
    mc.scale(2.0);
    """

    gtest.add(pre + code + post)

def test_mock_triangle(gtest):
    pre = """
    mock_cell mc;
    mc.fill_reference_triangle(2);
    double * coordinate_dofs = mc.coordinate_dofs;
    """

    post = """
    ASSERT_EQ(coordinate_dofs[0*mc.geometric_dimension + 0], 0.0);
    ASSERT_EQ(coordinate_dofs[0*mc.geometric_dimension + 1], 0.0);
    ASSERT_EQ(coordinate_dofs[1*mc.geometric_dimension + 0], 2.0);
    ASSERT_EQ(coordinate_dofs[1*mc.geometric_dimension + 1], 0.0);
    ASSERT_EQ(coordinate_dofs[2*mc.geometric_dimension + 0], 0.0);
    ASSERT_EQ(coordinate_dofs[2*mc.geometric_dimension + 1], 3.0);
    """

    code = """
    double factors[2] = { 2.0, 3.0 };
    mc.scale(factors);
    """

    gtest.add(pre + code + post)

def test_mock_tetrahedron(gtest):
    pre = """
    mock_cell mc;
    mc.fill_reference_tetrahedron(3);
    double * coordinate_dofs = mc.coordinate_dofs;
    """

    post = """
    ASSERT_EQ(coordinate_dofs[0*mc.geometric_dimension + 0], 2.0);
    ASSERT_EQ(coordinate_dofs[0*mc.geometric_dimension + 1], 3.0);
    ASSERT_EQ(coordinate_dofs[0*mc.geometric_dimension + 2], 4.0);
    ASSERT_EQ(coordinate_dofs[1*mc.geometric_dimension + 0], 3.0);
    ASSERT_EQ(coordinate_dofs[1*mc.geometric_dimension + 1], 3.0);
    ASSERT_EQ(coordinate_dofs[1*mc.geometric_dimension + 2], 4.0);
    ASSERT_EQ(coordinate_dofs[2*mc.geometric_dimension + 0], 2.0);
    ASSERT_EQ(coordinate_dofs[2*mc.geometric_dimension + 1], 4.0);
    ASSERT_EQ(coordinate_dofs[2*mc.geometric_dimension + 2], 4.0);
    ASSERT_EQ(coordinate_dofs[3*mc.geometric_dimension + 0], 2.0);
    ASSERT_EQ(coordinate_dofs[3*mc.geometric_dimension + 1], 3.0);
    ASSERT_EQ(coordinate_dofs[3*mc.geometric_dimension + 2], 5.0);
    """

    code = """
    double offset[3] = { 2.0, 3.0, 4.0 };
    mc.translate(offset);
    """

    gtest.add(pre + code + post)

def test_mock_quadrilateral(gtest):
    pre = """
    mock_cell mc;
    mc.fill_reference_quadrilateral(2);
    double * coordinate_dofs = mc.coordinate_dofs;
    """

    post = """
    ASSERT_EQ(coordinate_dofs[0*mc.geometric_dimension + 0], 2.0 + 2.0*0.0 + 3.0*0.0);
    ASSERT_EQ(coordinate_dofs[0*mc.geometric_dimension + 1], 3.0 + 4.0*0.0 + 5.0*0.0);
    ASSERT_EQ(coordinate_dofs[1*mc.geometric_dimension + 0], 2.0 + 2.0*1.0 + 3.0*0.0);
    ASSERT_EQ(coordinate_dofs[1*mc.geometric_dimension + 1], 3.0 + 4.0*1.0 + 5.0*0.0);
    ASSERT_EQ(coordinate_dofs[2*mc.geometric_dimension + 0], 2.0 + 2.0*1.0 + 3.0*1.0);
    ASSERT_EQ(coordinate_dofs[2*mc.geometric_dimension + 1], 3.0 + 4.0*1.0 + 5.0*1.0);
    ASSERT_EQ(coordinate_dofs[3*mc.geometric_dimension + 0], 2.0 + 2.0*0.0 + 3.0*1.0);
    ASSERT_EQ(coordinate_dofs[3*mc.geometric_dimension + 1], 3.0 + 4.0*0.0 + 5.0*1.0);
    """

    code = """
    double G[4] = { 2.0, 3.0,
                    4.0, 5.0 };
    double x[2] = { 2.0, 3.0 };
    mc.affine_map(G, x);
    """

    gtest.add(pre + code + post)

def test_mock_hexahedron(gtest):
    pre = """
    mock_cell mc;
    mc.fill_reference_hexahedron(3);
    double * coordinate_dofs = mc.coordinate_dofs;
    """

    post = """
    ASSERT_EQ(coordinate_dofs[0*mc.geometric_dimension + 0], 5.0 * (0.0 + 2.0));
    ASSERT_EQ(coordinate_dofs[0*mc.geometric_dimension + 1], 6.0 * (0.0 + 3.0));
    ASSERT_EQ(coordinate_dofs[0*mc.geometric_dimension + 2], 7.0 * (0.0 + 4.0));
    ASSERT_EQ(coordinate_dofs[1*mc.geometric_dimension + 0], 5.0 * (1.0 + 2.0));
    ASSERT_EQ(coordinate_dofs[1*mc.geometric_dimension + 1], 6.0 * (0.0 + 3.0));
    ASSERT_EQ(coordinate_dofs[1*mc.geometric_dimension + 2], 7.0 * (0.0 + 4.0));
    ASSERT_EQ(coordinate_dofs[2*mc.geometric_dimension + 0], 5.0 * (1.0 + 2.0));
    ASSERT_EQ(coordinate_dofs[2*mc.geometric_dimension + 1], 6.0 * (1.0 + 3.0));
    ASSERT_EQ(coordinate_dofs[2*mc.geometric_dimension + 2], 7.0 * (0.0 + 4.0));
    // ...
    ASSERT_EQ(coordinate_dofs[7*mc.geometric_dimension + 0], 5.0 * (0.0 + 2.0));
    ASSERT_EQ(coordinate_dofs[7*mc.geometric_dimension + 1], 6.0 * (1.0 + 3.0));
    ASSERT_EQ(coordinate_dofs[7*mc.geometric_dimension + 2], 7.0 * (1.0 + 4.0));
    """

    code = """
    double offset[3] = { 2.0, 3.0, 4.0 };
    double factors[3] = { 5.0, 6.0, 7.0 };
    mc.translate(offset);
    mc.scale(factors);
    """

    gtest.add(pre + code + post)




# -*- coding: utf-8 -*-
# Code generation format strings for UFC (Unified Form-assembly Code) version 2017.1.0
# This code is released into the public domain.
#
# The FEniCS Project (http://www.fenicsproject.org/) 2006-2017

cell_integral_combined = """
class %(classname)s: public ufc::cell_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s) : ufc::cell_integral()%(initializer_list)s
  {
%(constructor)s
  }

  ~%(classname)s() override
  {
%(destructor)s
  }

  const std::vector<bool> & enabled_coefficients() const final override
  {
%(enabled_coefficients)s
  }

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs,
                       int cell_orientation) const final override
  {
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
  }

};
"""

cell_integral_header = """
class %(classname)s: public ufc::cell_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s);

  ~%(classname)s() override;

  const std::vector<bool> & enabled_coefficients() const final override;

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs,
                       int cell_orientation) const final override;

};
"""

cell_integral_implementation = """
%(classname)s::%(classname)s(%(constructor_arguments)s) : ufc::cell_integral()%(initializer_list)s
{
%(constructor)s
}

%(classname)s::~%(classname)s()
{
%(destructor)s
}

const std::vector<bool> & %(classname)s::enabled_coefficients() const
{
%(enabled_coefficients)s
}

void %(classname)s::tabulate_tensor(double * A,
                                    const double * const * w,
                                    const double * coordinate_dofs,
                                    int cell_orientation) const
{
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
}
"""

exterior_facet_integral_combined = """
class %(classname)s: public ufc::exterior_facet_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s) : ufc::exterior_facet_integral()%(initializer_list)s
  {
%(constructor)s
  }

  ~%(classname)s() override
  {
%(destructor)s
  }

  const std::vector<bool> & enabled_coefficients() const final override
  {
%(enabled_coefficients)s
  }

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs,
                       std::size_t facet,
                       int cell_orientation) const final override
  {
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
  }

};
"""

exterior_facet_integral_header = """
class %(classname)s: public ufc::exterior_facet_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s);

  ~%(classname)s() override;

  const std::vector<bool> & enabled_coefficients() const final override;

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs,
                       std::size_t facet,
                       int cell_orientation) const final override;

};
"""

exterior_facet_integral_implementation = """
%(classname)s::%(classname)s(%(constructor_arguments)s) : ufc::exterior_facet_integral()%(initializer_list)s
{
%(constructor)s
}

%(classname)s::~%(classname)s()
{
%(destructor)s
}

const std::vector<bool> & %(classname)s::enabled_coefficients() const
{
%(enabled_coefficients)s
}

void %(classname)s::tabulate_tensor(double * A,
                                    const double * const * w,
                                    const double * coordinate_dofs,
                                    std::size_t facet,
                                    int cell_orientation) const
{
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
}
"""

interior_facet_integral_combined = """
class %(classname)s: public ufc::interior_facet_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s) : ufc::interior_facet_integral()%(initializer_list)s
  {
%(constructor)s
  }

  ~%(classname)s() override
  {
%(destructor)s
  }

  const std::vector<bool> & enabled_coefficients() const final override
  {
%(enabled_coefficients)s
  }

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs_0,
                       const double * coordinate_dofs_1,
                       std::size_t facet_0,
                       std::size_t facet_1,
                       int cell_orientation_0,
                       int cell_orientation_1) const final override
  {
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
  }

};
"""

interior_facet_integral_header = """
class %(classname)s: public ufc::interior_facet_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s);

  ~%(classname)s() override;

  const std::vector<bool> & enabled_coefficients() const final override;

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs_0,
                       const double * coordinate_dofs_1,
                       std::size_t facet_0,
                       std::size_t facet_1,
                       int cell_orientation_0,
                       int cell_orientation_1) const final override;

};
"""

interior_facet_integral_implementation = """
%(classname)s::%(classname)s(%(constructor_arguments)s) : ufc::interior_facet_integral()%(initializer_list)s
{
%(constructor)s
}

%(classname)s::~%(classname)s()
{
%(destructor)s
}

const std::vector<bool> & %(classname)s::enabled_coefficients() const
{
%(enabled_coefficients)s
}

void %(classname)s::tabulate_tensor(double * A,
                                    const double * const * w,
                                    const double * coordinate_dofs_0,
                                    const double * coordinate_dofs_1,
                                    std::size_t facet_0,
                                    std::size_t facet_1,
                                    int cell_orientation_0,
                                    int cell_orientation_1) const
{
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
}
"""

vertex_integral_combined = """
class %(classname)s: public ufc::vertex_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s) : ufc::vertex_integral()%(initializer_list)s
  {
%(constructor)s
  }

  ~%(classname)s() override
  {
%(destructor)s
  }

  const std::vector<bool> & enabled_coefficients() const final override
  {
%(enabled_coefficients)s
  }

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs,
                       std::size_t vertex,
                       int cell_orientation) const final override
  {
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
  }

};
"""

vertex_integral_header = """
class %(classname)s: public ufc::vertex_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s);

  ~%(classname)s() override;

  const std::vector<bool> & enabled_coefficients() const final override;

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs,
                       std::size_t vertex,
                       int cell_orientation) const final override;

};
"""

vertex_integral_implementation = """
%(classname)s::%(classname)s(%(constructor_arguments)s) : ufc::vertex_integral()%(initializer_list)s
{
%(constructor)s
}

%(classname)s::~%(classname)s()
{
%(destructor)s
}

const std::vector<bool> & %(classname)s::enabled_coefficients() const
{
%(enabled_coefficients)s
}

void %(classname)s::tabulate_tensor(double * A,
                                    const double * const * w,
                                    const double * coordinate_dofs,
                                    std::size_t vertex,
                                    int cell_orientation) const
{
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
}
"""

custom_integral_combined = """\
class %(classname)s: public ufc::custom_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s) : ufc::custom_integral()%(initializer_list)s
  {
%(constructor)s
  }

  ~%(classname)s() override
  {
%(destructor)s
  }

  const std::vector<bool> & enabled_coefficients() const final override
  {
%(enabled_coefficients)s
  }

  std::size_t num_cells() const final override
  {
%(num_cells)s
  }

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs,
                       std::size_t num_quadrature_points,
                       const double * quadrature_points,
                       const double * quadrature_weights,
                       const double * facet_normals,
                       int cell_orientation) const final override
  {
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
  }

};
"""

custom_integral_header = """
class %(classname)s: public ufc::custom_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s);

  ~%(classname)s() override;

  const std::vector<bool> & enabled_coefficients() const final override;

  std::size_t num_cells() const final override;

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs,
                       std::size_t num_quadrature_points,
                       const double * quadrature_points,
                       const double * quadrature_weights,
                       const double * facet_normals,
                       int cell_orientation) const final override;

};
"""

custom_integral_implementation = """
%(classname)s::%(classname)s(%(constructor_arguments)s) : ufc::custom_integral()%(initializer_list)s
{
%(constructor)s
}

%(classname)s::~%(classname)s()
{
%(destructor)s
}

const std::vector<bool> & %(classname)s::enabled_coefficients() const
{
%(enabled_coefficients)s
}

std::size_t %(classname)s::num_cells() const
{
%(num_cells)s
}

void %(classname)s::tabulate_tensor(double * A,
                                    const double * const * w,
                                    const double * coordinate_dofs,
                                    std::size_t num_quadrature_points,
                                    const double * quadrature_points,
                                    const double * quadrature_weights,
                                    const double * facet_normals,
                                    int cell_orientation) const
{
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
}
"""

cutcell_integral_combined = """
class %(classname)s: public ufc::cutcell_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s) : ufc::cutcell_integral()%(initializer_list)s
  {
%(constructor)s
  }

  ~%(classname)s() override
  {
%(destructor)s
  }

  const std::vector<bool> & enabled_coefficients() const final override
  {
%(enabled_coefficients)s
  }

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs,
                       std::size_t num_quadrature_points,
                       const double * quadrature_points,
                       const double * quadrature_weights,
                       int cell_orientation) const final override
  {
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
  }

};
"""

cutcell_integral_header = """
class %(classname)s: public ufc::cutcell_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s);

  ~%(classname)s() override;

  const std::vector<bool> & enabled_coefficients() const final override;

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs,
                       std::size_t num_quadrature_points,
                       const double * quadrature_points,
                       const double * quadrature_weights,
                       int cell_orientation) const final override;

};
"""

cutcell_integral_implementation = """
%(classname)s::%(classname)s(%(constructor_arguments)s) : ufc::cutcell_integral()%(initializer_list)s
{
%(constructor)s
}

%(classname)s::~%(classname)s()
{
%(destructor)s
}

const std::vector<bool> & %(classname)s::enabled_coefficients() const
{
%(enabled_coefficients)s
}

void %(classname)s::tabulate_tensor(double * A,
                                    const double * const * w,
                                    const double * coordinate_dofs,
                                    std::size_t num_quadrature_points,
                                    const double * quadrature_points,
                                    const double * quadrature_weights,
                                    int cell_orientation) const
{
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
}
"""

interface_integral_combined = """
class %(classname)s: public ufc::interface_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s) : ufc::interface_integral()%(initializer_list)s
  {
%(constructor)s
  }

  ~%(classname)s() override
  {
%(destructor)s
  }

  const std::vector<bool> & enabled_coefficients() const final override
  {
%(enabled_coefficients)s
  }

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs,
                       std::size_t num_quadrature_points,
                       const double * quadrature_points,
                       const double * quadrature_weights,
                       const double * facet_normals,
                       int cell_orientation) const final override
  {
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
  }

};
"""

interface_integral_header = """
class %(classname)s: public ufc::interface_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s);

  ~%(classname)s() override;

  const std::vector<bool> & enabled_coefficients() const final override;

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs,
                       std::size_t num_quadrature_points,
                       const double * quadrature_points,
                       const double * quadrature_weights,
                       const double * facet_normals,
                       int cell_orientation) const final override;

};
"""

interface_integral_implementation = """
%(classname)s::%(classname)s(%(constructor_arguments)s) : ufc::interface_integral()%(initializer_list)s
{
%(constructor)s
}

%(classname)s::~%(classname)s()
{
%(destructor)s
}

const std::vector<bool> & %(classname)s::enabled_coefficients() const
{
%(enabled_coefficients)s
}

void %(classname)s::tabulate_tensor(double * A,
                                    const double * const * w,
                                    const double * coordinate_dofs,
                                    std::size_t num_quadrature_points,
                                    const double * quadrature_points,
                                    const double * quadrature_weights,
                                    const double * facet_normals,
                                    int cell_orientation) const
{
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
}
"""

overlap_integral_combined = """
class %(classname)s: public ufc::overlap_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s) : ufc::overlap_integral()%(initializer_list)s
  {
%(constructor)s
  }

  ~%(classname)s() override
  {
%(destructor)s
  }

  const std::vector<bool> & enabled_coefficients() const final override
  {
%(enabled_coefficients)s
  }

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs,
                       std::size_t num_quadrature_points,
                       const double * quadrature_points,
                       const double * quadrature_weights,
                       int cell_orientation) const final override
  {
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
  }

};
"""

overlap_integral_header = """
class %(classname)s: public ufc::overlap_integral
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s);

  ~%(classname)s() override;

  const std::vector<bool> & enabled_coefficients() const final override;

  void tabulate_tensor(double * A,
                       const double * const * w,
                       const double * coordinate_dofs,
                       std::size_t num_quadrature_points,
                       const double * quadrature_points,
                       const double * quadrature_weights,
                       int cell_orientation) const final override;

};
"""

overlap_integral_implementation = """
%(classname)s::%(classname)s(%(constructor_arguments)s) : ufc::overlap_integral()%(initializer_list)s
{
%(constructor)s
}

%(classname)s::~%(classname)s()
{
%(destructor)s
}

const std::vector<bool> & %(classname)s::enabled_coefficients() const
{
%(enabled_coefficients)s
}

void %(classname)s::tabulate_tensor(double * A,
                                    const double * const * w,
                                    const double * coordinate_dofs,
                                    std::size_t num_quadrature_points,
                                    const double * quadrature_points,
                                    const double * quadrature_weights,
                                    int cell_orientation) const
{
%(tabulate_tensor_comment)s
%(tabulate_tensor)s
}
"""



# -*- coding: utf-8 -*-
"This file implements a class to represent a sum."

# Copyright (C) 2009-2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import six

from ufl.utils.sorting import sorted_by_key

# FFC modules.
from ffc.log import error
from ffc.cpp import format

# FFC quadrature modules.
from .symbolics import create_float
from .symbolics import create_product
from .symbolics import create_sum
from .symbolics import create_fraction
from .expr import Expr

from .floatvalue import FloatValue


class Sum(Expr):
    __slots__ = ("vrs", "_expanded", "_reduced")

    def __init__(self, variables):
        """Initialise a Sum object, it derives from Expr and contains the
        additional variables:

        vrs       - list, a list of variables.
        _expanded - object, an expanded object of self, e.g.,
                    self = 'x + x'-> self._expanded = 2*x (a product).
        _reduced  - object, a reduced object of self, e.g.,
                    self = '2*x + x*y'-> self._reduced = x*(2 + y) (a product).
        NOTE: self._prec = 3."""

        # Initialise value, list of variables, class, expanded and reduced.
        self.val = 1.0
        self.vrs = []
        self._prec = 3
        self._expanded = False
        self._reduced = False

        # Get epsilon
        EPS = format["epsilon"]
        # Process variables if we have any.
        if variables:
            # Loop variables and remove nested Sums and collect all floats in
            # 1 variable. We don't collect [x, x, x] into 3*x to avoid creating
            # objects, instead we do this when expanding the object.
            float_val = 0.0
            for var in variables:
                # Skip zero terms.
                if abs(var.val) < EPS:
                    continue
                elif var._prec == 0:  # float
                    float_val += var.val
                    continue
                elif var._prec == 3:  # sum
                    # Loop and handle variables of nested sum.
                    for v in var.vrs:
                        if abs(v.val) < EPS:
                            continue
                        elif v._prec == 0:  # float
                            float_val += v.val
                            continue
                        self.vrs.append(v)
                    continue
                self.vrs.append(var)

            # Only create new float if value is different from 0.
            if abs(float_val) > EPS:
                self.vrs.append(create_float(float_val))

        # If we don't have any variables the sum is zero.
        else:
            self.val = 0.0
            self.vrs = [create_float(0)]

        # Handle zero value.
        if not self.vrs:
            self.val = 0.0
            self.vrs = [create_float(0)]

        # Type is equal to the smallest type in both lists.
        self.t = min([v.t for v in self.vrs])

        # Sort variables, (for representation).
        self.vrs.sort()

        # Compute the representation now, such that we can use it directly
        # in the __eq__ and __ne__ methods (improves performance a bit, but
        # only when objects are cached).
        self._repr = "Sum([%s])" % ", ".join([v._repr for v in self.vrs])

        # Use repr as hash value.
        self._hash = hash(self._repr)

    # Print functions.
    def __str__(self):
        "Simple string representation which will appear in the generated code."
        # First add all the positive variables using plus, then add all
        # negative variables.
        s = format["add"]([str(v) for v in self.vrs if not v.val < 0]) +\
            "".join([str(v) for v in self.vrs if v.val < 0])
        # Group only if we have more that one variable.
        if len(self.vrs) > 1:
            return format["grouping"](s)
        return s

    # Binary operators.
    def __add__(self, other):
        "Addition by other objects."
        # Return a new sum
        return create_sum([self, other])

    def __sub__(self, other):
        "Subtract other objects."
        # Return a new sum
        return create_sum([self, create_product([FloatValue(-1), other])])

    def __mul__(self, other):
        "Multiplication by other objects."
        # If product will be zero.
        if self.val == 0.0 or other.val == 0.0:
            return create_float(0)

        # NOTE: We expect expanded sub-expressions with no nested operators.
        # Create list of new products using the '*' operator
        # TODO: Is this efficient?
        new_prods = [v * other for v in self.vrs]

        # Remove zero valued terms.
        # TODO: Can this still happen?
        new_prods = [v for v in new_prods if v.val != 0.0]

        # Create new sum.
        if not new_prods:
            return create_float(0)
        elif len(new_prods) > 1:
            # Expand sum to collect terms.
            return create_sum(new_prods).expand()
        # TODO: Is it necessary to call expand?
        return new_prods[0].expand()

    def __truediv__(self, other):
        "Division by other objects."
        # If division is illegal (this should definitely not happen).
        if other.val == 0.0:
            error("Division by zero.")

        # If fraction will be zero.
        if self.val == 0.0:
            return create_float(0)

        # NOTE: assuming that we get expanded variables.
        # If other is a Sum we can only return a fraction.
        # TODO: We could check for equal sums if Sum.__eq__ could be trusted.
        # As it is now (2*x + y) == (3*x + y), which works for the other things I do.
        # NOTE: Expect that other is expanded i.e., x + x -> 2*x which can be handled.
        # TODO: Fix (1 + y) / (x + x*y) -> 1 / x
        # Will this be handled when reducing operations on a fraction?
        if other._prec == 3:  # sum
            return create_fraction(self, other)

        # NOTE: We expect expanded sub-expressions with no nested operators.
        # Create list of new products using the '*' operator.
        # TODO: Is this efficient?
        new_fracs = [v / other for v in self.vrs]

        # Remove zero valued terms.
        # TODO: Can this still happen?
        new_fracs = [v for v in new_fracs if v.val != 0.0]

        # Create new sum.
        # TODO: No need to call expand here, using the '/' operator should have
        # taken care of this.
        if not new_fracs:
            return create_float(0)
        elif len(new_fracs) > 1:
            return create_sum(new_fracs)
        return new_fracs[0]

    __div__ = __truediv__

    # Public functions.
    def expand(self):
        "Expand all members of the sum."

        # If sum is already expanded, simply return the expansion.
        if self._expanded:
            return self._expanded

        # TODO: This function might need some optimisation.

        # Sort variables into symbols, products and fractions (add
        # floats directly to new list, will be handled later). Add
        # fractions if possible else add to list.
        new_variables = []
        syms = []
        prods = []
        # TODO: Rather than using '+', would it be more efficient to
        # collect the terms first?
        for var in self.vrs:
            exp = var.expand()
            # TODO: Should we also group fractions, or put this in a
            # separate function?
            if exp._prec in (0, 4):  # float or frac
                new_variables.append(exp)
            elif exp._prec == 1:  # sym
                syms.append(exp)
            elif exp._prec == 2:  # prod
                prods.append(exp)
            elif exp._prec == 3:  # sum
                for v in exp.vrs:
                    if v._prec in (0, 4):  # float or frac
                        new_variables.append(v)
                    elif v._prec == 1:  # sym
                        syms.append(v)
                    elif v._prec == 2:  # prod
                        prods.append(v)

        # Sort all variables in groups: [2*x, -7*x], [(x + y), (2*x +
        # 4*y)] etc.  First handle product in order to add symbols if
        # possible.
        prod_groups = {}
        for v in prods:
            if v.get_vrs() in prod_groups:
                prod_groups[v.get_vrs()] += v
            else:
                prod_groups[v.get_vrs()] = v

        sym_groups = {}
        # Loop symbols and add to appropriate groups.
        for v in syms:
            # First try to add to a product group.
            if (v,) in prod_groups:
                prod_groups[(v,)] += v
            # Then to other symbols.
            elif v in sym_groups:
                sym_groups[v] += v
            # Create a new entry in the symbols group.
            else:
                sym_groups[v] = v

        # Loop groups and add to new variable list.
        for k, v in sorted_by_key(sym_groups):
            new_variables.append(v)
        for k, v in sorted_by_key(prod_groups):
            new_variables.append(v)

        if len(new_variables) > 1:
            # Return new sum (will remove multiple instances of floats
            # during construction).
            self._expanded = create_sum(sorted(new_variables))
            return self._expanded
        elif new_variables:
            # If we just have one variable left, return it since it is
            # already expanded.
            self._expanded = new_variables[0]
            return self._expanded
        error("Where did the variables go?")

    def get_unique_vars(self, var_type):
        "Get unique variables (Symbols) as a set."
        # Loop all variables of self update the set.
        var = set()
        for v in self.vrs:
            var.update(v.get_unique_vars(var_type))
        return var

    def get_var_occurrences(self):
        """Determine the number of minimum number of times all variables
        occurs in the expression. Returns a dictionary of variables
        and the number of times they occur. x*x + x returns {x:1}, x +
        y returns {}.

        """
        # NOTE: This function is only used if the numerator of a
        # Fraction is a Sum.

        # Get occurrences in first expression.
        d0 = self.vrs[0].get_var_occurrences()
        for var in self.vrs[1:]:
            # Get the occurrences.
            d = var.get_var_occurrences()
            # Delete those variables in d0 that are not in d.
            for k, v in list(d0.items()):
                if k not in d:
                    del d0[k]
            # Set the number of occurrences equal to the smallest
            # number.
            for k, v in sorted_by_key(d):
                if k in d0:
                    d0[k] = min(d0[k], v)
        return d0

    def ops(self):
        "Return number of operations to compute value of sum."
        # Subtract one operation as it only takes n-1 ops to sum n
        # members.
        op = -1

        # Add the number of operations from sub-expressions.
        for v in self.vrs:
            #  +1 for the +/- symbol.
            op += v.ops() + 1
        return op

    def reduce_ops(self):
        "Reduce the number of operations needed to evaluate the sum."

        if self._reduced:
            return self._reduced
        # NOTE: Assuming that sum has already been expanded.
        # TODO: Add test for this and handle case if it is not.

        # TODO: The entire function looks expensive, can it be optimised?

        # TODO: It is not necessary to create a new Sum if we do not
        # have more than one Fraction.

        # First group all fractions in the sum.
        new_sum = _group_fractions(self)
        if new_sum._prec != 3:  # sum
            self._reduced = new_sum.reduce_ops()
            return self._reduced
        # Loop all variables of the sum and collect the number of
        # common variables that can be factored out.
        common_vars = {}
        for var in new_sum.vrs:
            # Get dictonary of occurrences and add the variable and
            # the number of occurrences to common dictionary.
            for k, v in sorted_by_key(var.get_var_occurrences()):
                if k in common_vars:
                    common_vars[k].append((v, var))
                else:
                    common_vars[k] = [(v, var)]

        # Determine the maximum reduction for each variable sorted as:
        # {(x*x*y, x*y*z, 2*y):[2, [y]]}.
        terms_reductions = {}
        for k, v in sorted_by_key(common_vars):
            # If the number of expressions that can be reduced is only
            # one there is nothing to be done.
            if len(v) > 1:
                # TODO: Is there a better way to compute the reduction
                # gain and the number of occurrences we should remove?

                # Get the list of number of occurences of 'k' in
                # expressions in 'v'.
                occurrences = [t[0] for t in v]

                # Determine the favorable number of occurences and an
                # estimate of the maximum reduction for current
                # variable.
                fav_occur = 0
                reduc = 0
                for i in set(occurrences):
                    # Get number of terms that has a number of
                    # occcurences equal to or higher than the current
                    # number.
                    num_terms = len([o for o in occurrences if o >= i])

                    # An estimate of the reduction in operations is:
                    # (number_of_terms - 1) * number_occurrences.
                    new_reduc = (num_terms - 1) * i
                    if new_reduc > reduc:
                        reduc = new_reduc
                        fav_occur = i

                # Extract the terms of v where the number of
                # occurrences is equal to or higher than the most
                # favorable number of occurrences.
                terms = sorted([t[1] for t in v if t[0] >= fav_occur])

                # We need to reduce the expression with the favorable
                # number of occurrences of the current variable.
                red_vars = [k] * fav_occur

                # If the list of terms is already present in the
                # dictionary, add the reduction count and the
                # variables.
                if tuple(terms) in terms_reductions:
                    terms_reductions[tuple(terms)][0] += reduc
                    terms_reductions[tuple(terms)][1] += red_vars
                else:
                    terms_reductions[tuple(terms)] = [reduc, red_vars]

        if terms_reductions:
            # Invert dictionary of terms.
            reductions_terms = dict([((v[0], tuple(v[1])), k) for k,
                                     v in six.iteritems(terms_reductions)])

            # Create a sorted list of those variables that give the
            # highest reduction.
            sorted_reduc_var = sorted(six.iterkeys(reductions_terms),
                                      reverse=True)

            # Create a new dictionary of terms that should be reduced,
            # if some terms overlap, only pick the one which give the
            # highest reduction to ensure that a*x*x + b*x*x + x*x*y +
            # 2*y -> x*x*(a + b + y) + 2*y NOT x*x*(a + b) + y*(2 +
            # x*x).
            reduction_vars = {}
            rejections = {}
            for var in sorted_reduc_var:
                terms = reductions_terms[var]
                if _overlap(terms, reduction_vars) or _overlap(terms,
                                                               rejections):
                    rejections[var[1]] = terms
                else:
                    reduction_vars[var[1]] = terms

            # Reduce each set of terms with appropriate variables.
            all_reduced_terms = []
            reduced_expressions = []
            for reduc_var, terms in sorted(six.iteritems(reduction_vars)):

                # Add current terms to list of all variables that have
                # been reduced.
                all_reduced_terms += list(terms)

                # Create variable that we will use to reduce the terms.
                reduction_var = None
                if len(reduc_var) > 1:
                    reduction_var = create_product(list(reduc_var))
                else:
                    reduction_var = reduc_var[0]

                # Reduce all terms that need to be reduced.
                reduced_terms = [t.reduce_var(reduction_var) for t in terms]

                # Create reduced expression.
                reduced_expr = None
                if len(reduced_terms) > 1:
                    # Try to reduce the reduced terms further.
                    reduced_expr = create_product([reduction_var,
                                                   create_sum(reduced_terms).reduce_ops()])
                else:
                    reduced_expr = create_product(reduction_var,
                                                  reduced_terms[0])

                # Add reduced expression to list of reduced
                # expressions.
                reduced_expressions.append(reduced_expr)

            # Create list of terms that should not be reduced.
            dont_reduce_terms = []
            for v in new_sum.vrs:
                if v not in all_reduced_terms:
                    dont_reduce_terms.append(v)

            # Create expression from terms that was not reduced.
            not_reduced_expr = None
            if dont_reduce_terms and len(dont_reduce_terms) > 1:
                # Try to reduce the remaining terms that were not
                # reduced at first.
                not_reduced_expr = create_sum(dont_reduce_terms).reduce_ops()
            elif dont_reduce_terms:
                not_reduced_expr = dont_reduce_terms[0]

            # Create return expression.
            if not_reduced_expr:
                self._reduced = create_sum(reduced_expressions + [not_reduced_expr])
            elif len(reduced_expressions) > 1:
                self._reduced = create_sum(reduced_expressions)
            else:
                self._reduced = reduced_expressions[0]

            return self._reduced

        # Return self if we don't have any variables for which we can
        # reduce the sum.
        self._reduced = self
        return self._reduced

    def reduce_vartype(self, var_type):
        """Reduce expression with given var_type. It returns a list of tuples
        [(found, remain)], where 'found' is an expression that only
        has variables of type == var_type. If no variables are found,
        found=(). The 'remain' part contains the leftover after
        division by 'found' such that: self = Sum([f*r for f,r in
        self.reduce_vartype(Type)]).

        """
        found = {}
        # Loop members and reduce them by vartype.
        for v in self.vrs:
            for f, r in v.reduce_vartype(var_type):
                if f in found:
                    found[f].append(r)
                else:
                    found[f] = [r]

        # Create the return value.
        returns = []
        for f, r in sorted_by_key(found):
            if len(r) > 1:
                # Use expand to group expressions.
                r = create_sum(r)
            elif r:
                r = r.pop()
            returns.append((f, r))
        return sorted(returns)


def _overlap(l, d):
    "Check if a member in list l is in the value (list) of dictionary d."
    for m in l:
        for k, v in sorted_by_key(d):
            if m in v:
                return True
    return False


def _group_fractions(expr):
    "Group Fractions in a Sum: 2/x + y/x -> (2 + y)/x."
    if expr._prec != 3:  # sum
        return expr

    # Loop variables and group those with common denominator.
    not_frac = []
    fracs = {}
    for v in expr.vrs:
        if v._prec == 4:  # frac
            if v.denom in fracs:
                fracs[v.denom][1].append(v.num)
                fracs[v.denom][0] += 1
            else:
                fracs[v.denom] = [1, [v.num], v]
            continue
        not_frac.append(v)
    if not fracs:
        return expr

    # Loop all fractions and create new ones using an appropriate
    # numerator.
    for k, v in sorted(six.iteritems(fracs)):
        if v[0] > 1:
            # TODO: Is it possible to avoid expanding the Sum?  I
            # think we have to because x/a + 2*x/a -> 3*x/a.
            not_frac.append(create_fraction(create_sum(v[1]).expand(), k))
        else:
            not_frac.append(v[2])

    # Create return value.
    if len(not_frac) > 1:
        return create_sum(not_frac)
    return not_frac[0]



# -*- coding: utf-8 -*-
# Copyright (C) 2011-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>

"""Algorithms for the representation phase of the form compilation."""

from ufl.classes import (ReferenceValue, ReferenceGrad, Grad,
                         CellAvg, FacetAvg,
                         PositiveRestricted, NegativeRestricted,
                         Indexed)
from ufl.corealg.multifunction import MultiFunction
from ufl.corealg.map_dag import map_expr_dag


modifier_precedence = [ReferenceValue, ReferenceGrad, Grad,
                       CellAvg, FacetAvg,
                       PositiveRestricted, NegativeRestricted,
                       Indexed]

modifier_precedence = { m._ufl_handler_name_: i for i, m in enumerate(modifier_precedence) }

# TODO: Move this to ufl?
# TODO: Add expr._ufl_modifier_precedence_ ? Add Terminal first and Operator last in the above list.


def balance_modified_terminal(expr):
    # NB! Assuminge e.g. grad(cell_avg(expr)) does not occur,
    # i.e. it is simplified to 0 immediately.

    if expr._ufl_is_terminal_:
        return expr

    assert expr._ufl_is_terminal_modifier_

    orig = expr

    # Build list of modifier layers
    layers = [expr]
    while not expr._ufl_is_terminal_:
        assert expr._ufl_is_terminal_modifier_
        expr = expr.ufl_operands[0]
        layers.append(expr)
    assert layers[-1] is expr
    assert expr._ufl_is_terminal_

    # Apply modifiers in order
    layers = sorted(layers[:-1], key=lambda e: modifier_precedence[e._ufl_handler_name_])
    for op in layers:
        ops = (expr,) + op.ufl_operands[1:]
        expr = op._ufl_expr_reconstruct_(*ops)

    # Preserve id if nothing has changed
    return orig if expr == orig else expr


class BalanceModifiers(MultiFunction):

    def expr(self, expr, *ops):
        return expr._ufl_expr_reconstruct_(*ops)

    def terminal(self, expr):
        return expr

    def _modifier(self, expr, *ops):
        return balance_modified_terminal(expr)

    reference_value = _modifier
    reference_grad = _modifier
    grad = _modifier
    cell_avg = _modifier
    facet_avg = _modifier
    positive_restricted = _modifier
    negative_restricted = _modifier


def balance_modifiers(expr):
    mf = BalanceModifiers()
    return map_expr_dag(mf, expr)



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import pytest
import time

# FFC modules
from ffc.quadrature.symbolics import *
from ffc.cpp import format, set_float_formatting
from ffc.parameters import FFC_PARAMETERS
set_float_formatting(FFC_PARAMETERS['precision'])


def testProduct():
    "Test simple product instance."

    f_0 = format["float"](0)
    f_1 = format["float"](1)
    f0 = FloatValue(-2.0)
    f1 = FloatValue(3.0)
    f2 = FloatValue(0)
    f3 = FloatValue(-1)
    f4 = FloatValue(1)
    f5 = FloatValue(-0.5)
    f6 = FloatValue(2.0)
    s0 = Symbol("x", BASIS)
    s1 = Symbol("y", GEO)
    s2 = Symbol("z", GEO)

    p0 = Product([])
    p1 = Product([s0])
    p2 = Product([s0, s1])
    p3 = Product([f1, s0, s1])
    p4 = Product([s0, f2, s2])
    p5 = Product([s0, f0, s1, f1, s2])
    p6 = Product([s0, f3, s1])
    p7 = Product([s0, f4, s1]).expand().reduce_ops()
    p8 = Product([s0, f0, s2, f5])
    p9 = Product([s0, s1])
    p10 = Product([p0, p1])
    p11 = Product([f5, f0])
    p12 = Product([f6, f5])
    p13 = Product([f6, f5]).expand()
    p14 = Product([f1, f2])
    p_tmp = Product([f1])
    p_tmp.expand()
    p15 = Product([p_tmp, s0])

    assert repr(p0) == "Product([FloatValue(%s)])" % f_0
    assert repr(p1) == "Product([Symbol('x', BASIS)])"
    assert repr(p3) == "Product([FloatValue(%s), Symbol('x', BASIS), Symbol('y', GEO)])"\
        % format["float"](3)
    assert repr(p6) == "Product([FloatValue(-%s), Symbol('x', BASIS), Symbol('y', GEO)])" % f_1
    assert repr(p7) == "Product([Symbol('x', BASIS), Symbol('y', GEO)])"
    assert repr(p8) == "Product([Symbol('x', BASIS), Symbol('z', GEO)])"
    assert str(p2) == 'x*y'
    assert str(p4) == '%s' % f_0
    assert str(p5) == '-%s*x*y*z' % format["float"](6)
    assert str(p6) == ' - x*y'
    assert str(p7) == 'x*y'
    assert str(p8) == 'x*z'
    assert str(p9) == 'x*y'
    assert p0.val == 0
    assert str(p10) == '%s' % f_0
    assert str(p11) == '%s' % f_1
    assert str(p12) == '-%s' % f_1
    assert str(p13) == '-%s' % f_1
    assert repr(p14) == "Product([FloatValue(%s)])" % f_0
    assert repr(p14.expand()) == "FloatValue(%s)" % f_0

    assert p1 == p1
    assert p1 != p7
    assert p4 != p3
    assert p2 == p9
    assert p2 != p3

    assert p0.ops() == 0
    assert p1.ops() == 0
    assert p2.ops() == 1
    assert p3.ops() == 2
    assert p4.ops() == 0
    assert p5.ops() == 3
    assert p6.ops() == 1
    assert p7.ops() == 1
    assert p8.ops() == 1
    assert p9.ops() == 1
    assert p10.ops() == 0
    assert p14.ops() == 0

    # Test hash
    l = [p3]
    d = {p3: 0}
    p10 = Product([f1, s0, s1])

    assert p3 in l
    assert p3 in d
    assert p10 in l
    assert p10 in d
    assert p2 not in l
    assert p2 not in d



# -*- coding: utf-8 -*-
"""
Tests of CRSArray data structure.
"""

from six.moves import xrange
from ffc.uflacs.analysis.crsarray import CRSArray


def test_crs_can_have_zero_element_rows():
    rcap, ecap = 3, 1
    A = CRSArray(rcap, ecap, int)
    for i in range(rcap):
        row = []
        A.push_row(row)
    assert A.num_elements == 0
    for i in range(rcap):
        row = []
        assert list(A[i]) == row


def test_crs_can_have_one_element_rows():
    rcap, ecap = 3, 3
    A = CRSArray(rcap, ecap, int)
    for i in range(rcap):
        row = [i]
        A.push_row(row)
    assert A.num_elements == rcap
    for i in range(rcap):
        row = [i]
        assert list(A[i]) == row


def test_crs_can_have_n_element_rows():
    rcap, ecap = 5, 25
    A = CRSArray(rcap, ecap, int)
    k = 0
    for i in range(rcap):
        row = [i + 2, i + 1] + [i] * i
        k += len(row)
        A.push_row(row)
    assert A.num_elements == k
    for i in range(rcap):
        row = [i + 2, i + 1] + [i] * i
        assert list(A[i]) == row
